from typing import Self, Any, TypeVar, Optional, TypeAlias, Container
from abc import ABC, abstractmethod

from copy import deepcopy, copy
from urllib.parse import urljoin

from presto.adict import adict

__all__ = "Request",

RequestT: TypeAlias = TypeVar("RequestT", bound="Request")
HandlerT: TypeAlias = TypeVar("HandlerT", bound="Request.Handler")


# noinspection PyPep8Naming
class Request(ABC):
    """"""
    APPEND_SLASH: bool = False

    __parent: Optional[RequestT] = None
    __handler: HandlerT = None
    __path: Optional[str] = None
    __params: adict
    __requests: dict[str, RequestT]

    Request: RequestT = NotImplemented

    # noinspection PyPep8Naming
    class Handler(ABC):
        class Response(ABC):
            _RAISE_FOR_STATUS: bool = True
            _RAISE_EXCEPT_FOR: Container = set()

            __handler: HandlerT
            __request: RequestT

            def __init__(self, hand: HandlerT, requ: RequestT):
                self.__handler = hand
                self.__request = requ
                self.raise_for_status()

            @property
            def handler(self) -> HandlerT:
                return self.__handler

            @property
            def prequest(self) -> RequestT:
                return self.__request

            def _should_raise_for_status(self, status_code: int) -> bool:
                return self._RAISE_FOR_STATUS and status_code not in self._RAISE_EXCEPT_FOR

            @abstractmethod
            def raise_for_status(self):
                raise NotImplementedError

            @property
            @abstractmethod
            def attr(self) -> adict:
                raise NotImplementedError

        @abstractmethod
        def __call__(self, requ: RequestT) -> Response:
            raise NotImplementedError

        async def A(self, requ: RequestT) -> Response:
            return self(requ)

    def __init__(
            self,
            *,
            parent: Optional[RequestT],
            path: Optional[str] = None,
            APPEND_SLASH: Optional[bool] = None,
            **kwds
    ):
        if parent is not None:
            self.__parent = parent

            if self.Request is NotImplemented:
                if parent.Request is NotImplemented:
                    raise TypeError("Request.Request must be defined for parent.")

                self.Request = parent.Request
        else:
            if self.Request is NotImplemented:
                raise TypeError("Request.Request must be defined for subclass.")

        self.__handler = self.Request.Handler()

        self.APPEND_SLASH = APPEND_SLASH if APPEND_SLASH is not None else self.APPEND_SLASH
        self.__path__ = path

        self.__params = adict(self.__clean_params__(getattr(self, "__params__", {})))
        self.__merge__(**kwds)

        self.__requests = dict()

    def __copy__(self) -> Self:
        this: Self = self.__class__.__new__(self.__class__)
        this.__parent = self.__parent
        this.__handler = self.__handler
        this.__path = self.__path
        this.APPEND_SLASH = self.APPEND_SLASH
        this.__params = self.__params
        this.__requests = self.__requests
        return this

    def __deepcopy__(self, memo: Optional[dict[int, Any]]) -> Self:
        this: Self = self.__copy__()
        this.__parent = deepcopy(self.__parent, memo)
        this.__handler = deepcopy(self.__handler, memo)
        this.__params = deepcopy(self.__params, memo)
        this.__requests = deepcopy(self.__requests, memo)
        return this

    def __repr__(self):
        params = self.__merged__
        url = self.__url__

        if url:
            url = f"url={url!r}"

        if params:
            params = f"{', ' if url else ''}params={params!r}"

        name = self.__class__.__name__

        if name.startswith("_"):
            name = name[1:]

        return f"{name}({url}{params})"

    @property
    def __path__(self) -> Optional[str]:
        return self.__path

    @__path__.setter
    def __path__(self, path: str):
        self.__path = path + ("/" if self.APPEND_SLASH and path[-1:] != "/" else "")

    @property
    def __url__(self) -> str:
        if self.__parent is None:
            return self.__path

        return urljoin(self.__parent.__url__ + "/", self.__path)

    @property
    def __merged__(self) -> adict:
        if self.__parent is None:
            return self.__params

        return self.__parent.__merged__.__merged__(self.__params)

    def __merge__(self, **kwds) -> None:
        self.__params.__merge__(self.__clean_params__(kwds))

    def __handle__(self) -> Handler.Response:
        return self.__handler(deepcopy(self))

    @property
    async def A(self):
        return await self.__handler.A(self)

    def __getitem__(self, item) -> Self:
        return self.__getattr__(str(item))

    def __getattr__(self, name: str) -> Self:
        if "__" in name:
            raise AttributeError(name)  # Most likely an accidentally invalid internal attribute/method access.

        if (request := self.__requests.get(name)) is None:
            request = self.__requests[name] = self.Request(
                parent=self,
                path=name,
                APPEND_SLASH=self.APPEND_SLASH
            )

        elif request.__parent is not self:
            req = copy(request)
            req.__parent = self
            request = self.__requests[name] = req

        return request

    def __call__(self, **kwds) -> Self | Handler.Response:
        if not kwds:
            return self.__handle__()

        self.__merge__(**kwds)
        return self

    @classmethod
    def __clean_params__(cls, params: adict | dict) -> adict | dict:
        if "url" in params:
            raise ValueError("url is a reserved request attribute.")

        return params

    def __eq__(self, other) -> bool:
        if not isinstance(other, Request):
            return False

        return self.__url__ == other.__url__ and self.__merged__ == other.__merged__



from typing import Optional, TypeAlias, TypeVar, Any, Self

import requests

from presto.adict import adict
from presto.request import Request

__all__ = "Handler",

HandlerT: TypeAlias = TypeVar("HandlerT", bound="Handler")


class Handler(Request.Handler):
    """Base request handler."""

    __session: Optional[requests.Session] = None

    class Response(requests.Response, Request.Handler.Response):
        __attr: Optional[adict] = None

        def __init__(self, hand: HandlerT, requ: Request, resp: requests.Response):
            self.__dict__ = resp.__dict__
            self.__attr = None
            Request.Handler.Response.__init__(self, hand, requ)

        @property
        def attr(self):
            if self.__attr is None:
                if self.status_code == 200 and \
                        self.headers.get("content-type").startswith("application/json"):
                    self.__attr = adict(self.json())
            return self.__attr

        def raise_for_status(self):
            if self._should_raise_for_status(self.status_code):
                super().raise_for_status()

    @classmethod
    def __request__(cls, requ: Request) -> tuple[str, str, dict[str, Any]]:
        url = requ.__url__
        rreq = requ.__merged__
        method = rreq.pop("method")
        return method, url, rreq

    def __call__(self, requ: Request) -> Response:
        method, url, rreq = self.__request__(requ)
        return self.Response(
            hand=self,
            requ=requ,
            resp=self.session.request(
                method=method,
                url=url,
                **rreq
            )
        )

    @property
    def session(self) -> requests.Session:
        if self.__session is None:
            self.__session = requests.Session()
        return self.__session

    def __copy__(self) -> Self:
        this: Self = self.__class__.__new__(self.__class__)
        this.__session = self.__session
        return this

    def __deepcopy__(self, memo: dict[int, Any]) -> Self:
        this = self.__copy__()
        this.__session = None
        return this



from typing import TypeAlias, TypeVar, Self, Any

from attrdict import AttrDict

__all__ = "Adict", "adict",

Adict: TypeAlias = TypeVar("Adict", bound="adict | dict", covariant=True)


# noinspection PyPep8Naming
class adict(AttrDict):
    def __repr__(self: Self) -> str:
        return f"{self.__class__.__name__}({', '.join(f'{str(key)}={value!r}' for key, value in self.items())})"

    def __merge__(self: Adict, other: Adict) -> Adict:
        for key, value in other.items():
            if key in self and isinstance(self[key], dict) and isinstance(value, dict):
                self[key] = adict.__merge__(self[key], value)
            else:
                self[key] = value

        return self

    def __merged__(self: Adict, other: Adict) -> Adict:
        this = adict(self) if isinstance(self, adict) else dict(self)
        return adict.__merge__(this, other)

    def __delitem__(self: Self, key: Any) -> Self:

        for key, val in tuple(self.items()):
            if isinstance(val, (adict, dict)):
                adict.__delitem__(val, key)

            super().__delitem__(key)

        return self




from typing import Optional, Type, Self, TypeAlias, TypeVar

from copy import copy, deepcopy

from presto.adict import adict
from presto.request import Request
from presto.handler import Handler

__all__ = "Presto",

ClientT: TypeAlias = TypeVar("ClientT", bound="Client")


# noinspection PyPep8Naming
class Presto(Request):

    Client: Type[ClientT]

    class Request(Request):
        class Handler(Handler):
            class Response(Handler.Response):
                ...

    __params__: adict = adict(
        method="GET",
        headers=dict(
            Accept="application/json",
        ),
    )

    def __init__(
            self,
            *,
            url: str,
            RequestType: Optional[Type[Request]] = None,
            APPEND_SLASH: Optional[bool] = None,
            **kwds
    ):
        self.Request = RequestType or self.Request

        super().__init__(
            parent=None,
            path=url,
            APPEND_SLASH=APPEND_SLASH,
            **kwds
        )

    def __copy__(self) -> Self:
        this: Self = super().__copy__()
        this.Request = self.Request
        return this

    @property
    def request(self) -> Request:
        return super().__getattr__("")

    def __call__(self, url: Optional[str] = None, **kwds) -> Self | Request.Handler.Response:
        if url is not None:
            if not kwds:
                this = copy(self)
                this.__path__ = url
                return this

            this = deepcopy(self)
            this.__path__ = url
            this.__merge__(**kwds)
            return this

        return super().__call__(**kwds)

    get = property(lambda self: self.request(method="GET"))
    post = property(lambda self: self.request(method="POST"))
    put = property(lambda self: self.request(method="PUT"))
    patch = property(lambda self: self.request(method="PATCH"))
    delete = property(lambda self: self.request(method="DELETE"))
    options = property(lambda self: self.request(method="OPTIONS"))
    head = property(lambda self: self.request(method="HEAD"))


class Client:
    """Very simple base class for Presto client API implementations."""

    P: Presto

    class Request(Presto.Request):
        class Handler(Presto.Request.Handler):
            class Response(Presto.Request.Handler.Response):
                _RAISE_EXCEPT_FOR = {404}

    # noinspection PyPep8Naming
    def __init__(
            self,
            *,
            url: str,
            PrestoType: Optional[Type[Presto]] = None,
            RequestType: Optional[Type[Request]] = None,
            APPEND_SLASH: Optional[bool] = None,
            **kwds
    ):
        self.P = (PrestoType or Presto)(
            url=url,
            RequestType=RequestType or self.Request,
            APPEND_SLASH=APPEND_SLASH,
        )

    @property
    def url(self):
        return self.P.__url__


Presto.Client = Client



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['presto']

package_data = \
{'': ['*']}

install_requires = \
['attrdict3>=2.0.2,<3.0.0', 'requests>=2.28.1,<3.0.0']

extras_require = \
{'async': ['httpx>=0.23.1,<0.24.0']}

setup_kwargs = {
    'name': 'presto-requests',
    'version': '1.3.1',
    'description': 'A modern object-oriented HTTP REST client.',
    'long_description': '# Presto! Requests\n\nAn object-oriented REST API client & requests extesion library.\n\nAs of version 1.0.0, this library can be considered stable, and significant\nchanges to the interface or new features are unlikely to be introduced.\n\n#### Note: This library requires Python 3.11.0rc1 or later.\n\n## Installation\n\n```bash\npip install presto-requests\n```\n```bash\npoetry add presto-requests\n```\n\n#### With async support using httpx:\n```bash\npip install presto-requests[async]\n```\n```bash\npoetry add presto-requests --extras async\n```\n\n### Concept:\n\nPresto! Requests is a library that extends the functionality of the requests library.\nIt provides a simple way to create a REST API client that is object-oriented and easy to use.\n\n### Example:\n\n```python\nfrom presto import Presto\nimport pprint\n\npresto = Presto("https://api.github.com")\n\nuser = presto.users.sitbon()  # == presto.users["sitbon"]()\n\nprint(f"User {user.attr.login} has {user.attr.public_repos} public repositories.")\n\npprint.pprint(user.json())\n```\n```shell\nUser sitbon has 15 public repositories.\n{\'avatar_url\': \'https://avatars.githubusercontent.com/u/1381063?v=4\',\n \'bio\': None,\n \'blog\': \'\',\n \'company\': None,\n \'created_at\': \'2012-01-26T04:25:21Z\',\n \'email\': None,\n \'events_url\': \'https://api.github.com/users/sitbon/events{/privacy}\',\n \'followers\': 7,\n \'followers_url\': \'https://api.github.com/users/sitbon/followers\',\n \'following\': 13,\n \'following_url\': \'https://api.github.com/users/sitbon/following{/other_user}\',\n \'gists_url\': \'https://api.github.com/users/sitbon/gists{/gist_id}\',\n \'gravatar_id\': \'\',\n \'hireable\': None,\n \'html_url\': \'https://github.com/sitbon\',\n \'id\': 1381063,\n \'location\': \'Portland, OR, USA\',\n \'login\': \'sitbon\',\n \'name\': \'Phillip Sitbon\',\n \'node_id\': \'MDQ6VXNlcjEzODEwNjM=\',\n \'organizations_url\': \'https://api.github.com/users/sitbon/orgs\',\n \'public_gists\': 4,\n \'public_repos\': 15,\n \'received_events_url\': \'https://api.github.com/users/sitbon/received_events\',\n \'repos_url\': \'https://api.github.com/users/sitbon/repos\',\n \'site_admin\': False,\n \'starred_url\': \'https://api.github.com/users/sitbon/starred{/owner}{/repo}\',\n \'subscriptions_url\': \'https://api.github.com/users/sitbon/subscriptions\',\n \'twitter_username\': None,\n \'type\': \'User\',\n \'updated_at\': \'2022-11-22T00:41:18Z\',\n \'url\': \'https://api.github.com/users/sitbon\'}\n\n```\n\n### Usage:\n\nEach dot in the path of the request is a new request object.\n\nCalling the object without any arguments will execute the request and return the response object.\n\nIndexing the object like a list is a convient way to extend the path to a new object for things\nlike id paths, e.g. `presto.note[1]()`.\n\nSpecifying keyword arguments will add them to the request as keyword arguments to requests.request(),\nand then return the current object for further chaining.\n\nThere are a few special top-level attributes that can be used to modify the request:\n`get`, `post`, `put`, `patch`, `delete`, `head`, `options`, and finally `request` which is\nan empty path component that can be used to indirectly modify existing top-level auto created request objects.\n\nAll of these top-level attributes are able to clone existing request attributes, to modify the path\nand parent parameters while using the same component path and parameters.\n\nFor example:\n\n```python\nfrom presto import Presto\n\n\npresto = Presto(url="http://127.0.0.1:8000", APPEND_SLASH=True)\n\nprint("presto:", presto)\n\napi = presto.api\n\nprint("api:", api)\n\nnote = api(headers={"X-User": "Testing"}).note\n\nprint("api.note:", api.note, "equal:", api.note == note)\n\nresp = api.note[4]()\n\nprint("headers:", resp.request.headers)\nprint("response:", resp)\nprint("note:", resp.attr)\n```\n```shell\npresto: Presto(url=\'http://127.0.0.1:8000/\', params=adict(method=\'GET\', headers={\'Accept\': \'application/json\'}))\napi: Request(url=\'http://127.0.0.1:8000/api/\', params=adict(method=\'GET\', headers={\'Accept\': \'application/json\'}))\napi.note: Request(url=\'http://127.0.0.1:8000/api/note/\', params=adict(method=\'GET\', headers={\'Accept\': \'application/json\', \'X-User\': \'Testing\'})) equal: True\nheaders: {\'User-Agent\': \'python-requests/2.28.1\', \'Accept-Encoding\': \'gzip, deflate\', \'Accept\': \'application/json\', \'Connection\': \'keep-alive\', \'X-User\': \'Testing\'}\nresponse: <Response [200]>\nnote: adict(user=\'\', id=4, url=\'http://127.0.0.1:8000/api/note/4/\', time=\'2022-12-02T19:26:09-0800\', note=\'Hello from the API!!\', collection={\'id\': 3, \'url\': \'http://127.0.0.1:8000/api/note/coll/3/\', \'name\': \'Public\', \'public\': True, \'notes\': 1})\n```\n\n`response.attr` is an `adict` instance, which is a dictionary that can be accessed as attributes.\nIt contains the JSON-decoded content of a response, if any.\n\n`APPEND_SLASH` is meant to be client implementation-specific, e.g. for a Django Rest Framework client, one would\ntypically set `Presto.APPEND_SLASH = True` or inherit from `Presto` in a pre-defined API client class.\n\n# Async Support\n\nVersion 1.0.0 adds support for async requests using the `httpx` library.\n\nThe usage is the same as the synchronous version except when it comes to executing requests.\n\nInstead of calling `()` an object to execute the request, the .A attribute is awaited.\nSee the example below.\n\n```python\nimport asyncio\nfrom presto.asynco import AsyncPresto as Presto\n\n\nasync def main():\n    presto = Presto(url="http://127.0.0.1:8000", APPEND_SLASH=True)\n\n    print("presto:", presto)\n\n    api = presto.api\n\n    print("api:", api)\n\n    note = api(headers={"X-User": "Testing"}).note\n\n    print("api.note:", api.note, "equal:", api.note == note)\n\n    resp = await api.note[4].A\n\n    print("headers:", resp.request.headers)\n    print("response:", resp)\n    print("note:", resp.attr)\n\nif __name__ == "__main__":\n    asyncio.run(main())\n```\n```shell\npresto: AsyncPresto(url=\'http://127.0.0.1:8000/\', params=adict(method=\'GET\', headers={\'Accept\': \'application/json\'}))\napi: Request(url=\'http://127.0.0.1:8000/api/\', params=adict(method=\'GET\', headers={\'Accept\': \'application/json\'}))\napi.note: Request(url=\'http://127.0.0.1:8000/api/note/\', params=adict(method=\'GET\', headers={\'Accept\': \'application/json\', \'X-User\': \'Testing\'})) equal: True\nheaders: Headers({\'host\': \'127.0.0.1:8000\', \'accept-encoding\': \'gzip, deflate\', \'connection\': \'keep-alive\', \'user-agent\': \'python-httpx/0.23.1\', \'accept\': \'application/json\', \'x-user\': \'Testing\'})\nresponse: <Response [200 OK]>\nnote: adict(user=\'\', id=4, url=\'http://127.0.0.1:8000/api/note/4/\', time=\'2022-12-02T19:26:09-0800\', note=\'Hello from the API!!\', collection={\'id\': 3, \'url\': \'http://127.0.0.1:8000/api/note/coll/3/\', \'name\': \'Public\', \'public\': True, \'notes\': 1})\n```',
    'author': 'Phillip Sitbon',
    'author_email': 'phillip.sitbon@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/sitbon/presto',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.11.0rc1,<4.0.0',
}


setup(**setup_kwargs)



from typing import Optional, TypeAlias, TypeVar, Self, Type

import httpx

from presto.presto import Presto

__all__ = "AsyncPresto",

HandlerT: TypeAlias = TypeVar("HandlerT", bound="AsyncPresto.Request.Handler")
ClientT: TypeAlias = TypeVar("ClientT", bound="Client")


class AsyncPresto(Presto):

    Client: Type[ClientT]

    class Request(Presto.Request):
        class Handler(Presto.Request.Handler):

            _client: Optional[httpx.AsyncClient] = None

            class Response(httpx.Response, Presto.Request.Handler.Response):
                def __init__(self, hand: HandlerT, requ: Presto.Request, resp: httpx.Response):
                    Presto.Request.Handler.Response.__init__(self, hand, requ, resp)

            @property
            def client(self) -> httpx.AsyncClient:
                if self._client is None:
                    self._client = httpx.AsyncClient()
                return self._client

            async def A(self, requ: Presto.Request) -> Response:
                method, url, params = self.__request__(requ)
                return self.Response(self, requ, await self.client.request(method, url, **params))

        def __call__(self, **kwds) -> Self:
            if not kwds:
                return self.A

            return super().__call__(**kwds)

    def __call__(self, url: Optional[str] = None, **kwds) -> Self:
        if url is None and not kwds:
            return self.A

        return super().__call__(url=url, **kwds)


# noinspection PyPep8Naming
class Client(AsyncPresto.Client):
    def __init__(
            self,
            *,
            url: str,
            RequestType: Optional[Type[AsyncPresto.Request]] = None,
            APPEND_SLASH: Optional[bool] = None,
            **kwds
    ):
        super().__init__(
            url=url,
            RequestType=RequestType or self.Request,
            APPEND_SLASH=APPEND_SLASH,
            **kwds
        )


AsyncPresto.Client = Client
